import logging.config
import os
from configparser import ConfigParser
from pathlib import Path
import requests
from infoworks.core.iw_authentication import is_token_valid, get_bearer_token
from infoworks.sdk import local_configurations
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from infoworks.sdk.utils import IWUtils
import urllib3
urllib3.disable_warnings()


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = local_configurations.REQUEST_TIMEOUT_IN_SEC
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def initialise_http_client():
    retries = Retry(total=local_configurations.MAX_RETRIES, backoff_factor=1,
                    status_forcelist=[429, 500, 502, 503, 504])
    http = requests.Session()
    adapter = TimeoutHTTPAdapter(max_retries=retries)
    http.mount("https://", adapter)
    http.mount("http://", adapter)
    return http


class BaseClient(object):
    def __init__(self):
        # print("Inside BaseClient")
        self.client_config = {
            'protocol': None,
            'ip': None,
            'port': None,
            'refresh_token': None,
            'bearer_token': None,
            'default_environment_id': None,
            'default_storage_id': None,
            'default_compute_id': None
        }
        self.secrets_config = {"custom_secrets_read": False}
        self.mappings = {}
        self.http = initialise_http_client()
        log_path = Path(local_configurations.LOG_LOCATION)
        if os.path.isdir(log_path.parent.absolute()):
            logging.basicConfig(filename=local_configurations.LOG_LOCATION, filemode='w',
                                format='%(asctime)s - %(module)s - %(pathname)s - %(lineno)d - %(levelname)s - %(message)s',
                                level=logging.DEBUG, datefmt='%Y-%m-%d %H:%M:%S')
        else:
            logging.basicConfig(
                format='%(asctime)s - %(module)s - %(pathname)s - %(lineno)d - %(levelname)s - %(message)s',
                level=logging.DEBUG, datefmt='%Y-%m-%d %H:%M:%S')
        self.logger = logging.getLogger('infoworks_sdk_logs')

    def initialize_client(self, protocol=None, ip=None, port=None, ):
        """
        initialize the client
        :param protocol: protocol to be used for server call
        :type protocol: String
        :param ip: Client IP
        :type ip: String
        :param port: Client Port
        :type port: String
        """

        if all(v is not None for v in [protocol, ip, port]):
            self.client_config['protocol'] = protocol
            self.client_config['ip'] = ip
            self.client_config['port'] = port
        else:
            logging.error(ValueError("Pass valid values"))
            raise ValueError("Pass valid values")

    def initialize_client_with_defaults(self, protocol, ip, port, refresh_token, default_environment_id=None,
                                        default_storage_id=None, default_compute_id=None):
        """
        initializes the client and a user with given configuration
        :param protocol: protocol to be used for server call
        :type protocol: String
        :param ip: ip address of the server
        :type ip: String
        :param port: port on which the rest service resides
        :type port: String
        :param refresh_token: refresh_token of the user
        :type refresh_token: String
        :param default_compute_id: Pass the default compute id to be used for all the artifacts to be created
        :param default_storage_id: Pass the default storage id to be used for all the artifacts to be created
        :param default_environment_id: Pass the default environment id to be used for all the artifacts to be created
        """
        self.client_config['refresh_token'] = refresh_token
        self.client_config['bearer_token'] = get_bearer_token(protocol, ip, port, refresh_token)
        self.initialize_client(protocol, ip, port)
        self.client_config[
            'default_environment_id'] = default_environment_id if default_environment_id is not None else None
        self.client_config['default_storage_id'] = default_storage_id if default_storage_id is not None else None
        self.client_config['default_compute_id'] = default_compute_id if default_compute_id is not None else None

    def get_configurations(self):
        """
        returns client configurations
        :return: configuration dictionary
        """
        return self.client_config

    def regenerate_bearer_token_if_needed(self, headers):
        if not is_token_valid(self.client_config, self.http):
            self.client_config['bearer_token'] = get_bearer_token(protocol=self.client_config["protocol"],
                                                                  ip=self.client_config["ip"],
                                                                  port=self.client_config["port"],
                                                                  refresh_token=self.client_config["refresh_token"])
            headers = IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
        return headers

    def call_api(self, method, url, headers=None, data=None):
        # headers = self.regenerate_bearer_token_if_needed(headers)
        if method.upper() == "GET":
            self.logger.info(f"Calling {url}")
            response = self.http.get(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                     verify=False, data=data)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.get(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                     verify=False, data=data)
            else:
                return response
        elif method.upper() == "POST":
            self.logger.info(f"Calling {url}")
            response = self.http.post(url, headers=headers, json=data,
                                      timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.post(url, headers=headers, json=data,
                                      timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            else:
                return response
        elif method.upper() == "PUT":
            self.logger.info(f"Calling {url}")
            response = self.http.put(url, headers=headers, json=data,
                                     timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.put(url, headers=headers, json=data,
                                     timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            else:
                return response
        elif method.upper() == "PATCH":
            self.logger.info(f"Calling {url}")
            response = self.http.patch(url, headers=headers, json=data,
                                       timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.put(url, headers=headers, json=data,
                                     timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            else:
                return response
        elif method.upper() == "DELETE":
            self.logger.info(f"Calling {url}")
            response = self.http.delete(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                        verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.delete(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                        verify=False)
            else:
                return response

    def get_mappings_from_config_file(self, ini_config_file_path):
        config = ConfigParser()
        config.optionxform = str
        config.read(ini_config_file_path)
        if "environment" in config:
            default_environment_id = config.get("environment", "environment_id")
            default_storage_id = config.get("environment", "storage_id")
            default_interactive_id = config.get("environment", "interactive_id")
            self.client_config["default_environment_id"] = default_environment_id
            self.client_config["default_storage_id"] = default_storage_id
            self.client_config["default_compute_id"] = default_interactive_id
        if "environment_mappings" in config:
            environment_mappings = dict(config.items("environment_mappings"))
            self.mappings["environment_mappings"] = environment_mappings
        if "storage_mappings" in config:
            storage_mappings = dict(config.items("storage_mappings"))
            self.mappings["storage_mappings"] = storage_mappings
        if "compute_mappings" in config:
            compute_mappings = dict(config.items("compute_mappings"))
            self.mappings["compute_mappings"] = compute_mappings
        if "gcp_details" in config:
            gcp_details = dict(config.items("gcp_details"))
            self.mappings["gcp_details"] = gcp_details
        if "gcp_project_id_mappings" in config:
            gcp_project_id_mappings = dict(config.items("gcp_project_id_mappings"))
            self.mappings["gcp_project_id_mappings"] = gcp_project_id_mappings
        if "service_json_mappings" in config:
            service_json_mappings = dict(config.items("service_json_mappings"))
            self.mappings["service_json_mappings"] = service_json_mappings
        if "table_group_compute_mappings" in config:
            table_group_compute_mappings = dict(config.items("table_group_compute_mappings"))
            self.mappings["table_group_compute_mappings"] = table_group_compute_mappings
        if "kms" in config:
            kms = dict(config.items("kms"))
            self.mappings["kms"] = kms
        for section in config.sections():
            self.mappings[section] = dict(config.items(section))

    def _topological_sort_grouping(self, g):
        # copy the graph
        _g = g.copy()
        res = []
        # while _g is not empty
        while _g:
            zero_indegree = [v for v, d in _g.in_degree() if d == 0]
            res.append(zero_indegree)
            _g.remove_nodes_from(zero_indegree)
        return res




import os
from infoworks.sdk.cicd.download_configurations.utils import Utils
from infoworks.sdk.base_client import BaseClient


class DownloadWorkflow(BaseClient):
    def __init__(self):
        super(DownloadWorkflow, self).__init__()

    def cicd_get_workflowconfig_dumps(self, workflow_ids, config_file_dump_path, files_overwrite=True,
                                      serviceaccountemail="admin@infoworks.io",
                                      replace_words=""):
        if not os.path.exists(os.path.join(config_file_dump_path, "modified_files")):
            os.makedirs(os.path.join(config_file_dump_path, "modified_files"))
        if not os.path.exists(os.path.join(config_file_dump_path, "workflow")):
            os.makedirs(os.path.join(config_file_dump_path, "workflow"))
        utils_obj = Utils(serviceaccountemail)
        target_file_path = os.path.join(os.path.join(config_file_dump_path, "modified_files"), "workflow.csv")
        if files_overwrite:
            open(target_file_path, 'w').close()
            mode = "a"
        else:
            mode = "a"
        f = open(target_file_path, mode)
        for workflow_id in workflow_ids:
            try:
                json_obj = {"entity_id": workflow_id, "entity_type": "workflow"}
                domain_id = utils_obj.get_domain_id(self, json_obj)
                if domain_id:
                    filename, configuration_obj = utils_obj.dump_to_file(self, "workflow", domain_id, workflow_id,
                                                                         replace_words, config_file_dump_path)
                    if filename is not None:
                        f.write(filename)
                        f.write("\n")
            except Exception as e:
                self.logger.error(f"Unable to export configurations for workflow {workflow_id} due to {str(e)}")
                print(f"Unable to export configurations for workflow {workflow_id} due to {str(e)}")
        f.close()



class AuthenticationError(Exception):
    def __init__(self, message):
        self.message = message
        super(AuthenticationError, self).__init__(self.message)


class SourceError(Exception):
    def __init__(self, message):
        self.message = message
        super(SourceError, self).__init__(self.message)


class PipelineError(Exception):
    def __init__(self, message):
        self.message = message
        super(PipelineError, self).__init__(self.message)


class WorkflowError(Exception):
    def __init__(self, message):
        self.message = message
        super(WorkflowError, self).__init__(self.message)


class DomainError(Exception):
    def __init__(self, message):
        self.message = message
        super(DomainError, self).__init__(self.message)


class AdminError(Exception):
    def __init__(self, message):
        self.message = message
        super(AdminError, self).__init__(self.message)


class JobsError(Exception):
    def __init__(self, message):
        self.message = message
        super(JobsError, self).__init__(self.message)


class GenericError(Exception):
    def __init__(self, message):
        self.message = message
        super(GenericError, self).__init__(self.message)



import requests

from infoworks.error import DomainError
from infoworks.sdk import url_builder
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.local_configurations import Response, ErrorCode
from infoworks.sdk.generic_response import GenericResponse
from infoworks.sdk.url_builder import get_parent_entity_url
from infoworks.sdk.utils import IWUtils


class DomainClient(BaseClient):
    def __init__(self):
        super(DomainClient, self).__init__()

    def create_domain(self, config_body):
        """
        Function to create domain
        :param config_body: JSON payload for the domain creation
        :type config_body: JSON dict
        config_body_example = {
         "name": "Abhi",
         "description": "Example Domain",
         "environment_id": "d60a47e2c8438a3d6daf7958"
        }
        :return: response dict
        """
        try:
            create_domain_url = url_builder.create_domain_url(self.client_config)
            response = self.call_api("POST", create_domain_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response, job_id=None)
        except Exception as e:
            self.logger.error(f"Failed to create domain" + str(e))
            raise DomainError(f"Failed to create domain" + str(e))

    def list_domains(self, params=None):
        """
        Function to list domains
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_domains = url_builder.create_domain_url(
            self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
        domains_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_domains,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    domains_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
                response["result"] = domains_list
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    response="Failed to get list of domains"
                                                    )
        except Exception as e:
            self.logger.error("Error in listing domains")
            raise DomainError("Error in listing domains" + str(e))

    def get_domain_details(self, domain_id):
        """
        Function to get the domain details
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :return: response dict
        """
        if domain_id is None:
            raise Exception("domain_id cannot be None")
        try:
            url_to_get_domain_details = url_builder.create_domain_url(self.client_config) + f"/{domain_id}"
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_to_get_domain_details,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               ).content)
            result = response.get('result', False)
            if not result:
                self.logger.error(f"Failed to get the domain details for {domain_id} ")
                return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f"Failed to get the domain details for {domain_id} ",
                                                    job_id=None, response=response)
            else:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error(f"Failed to get the domain details for {domain_id} " + str(e))
            raise DomainError(f"Failed to get the domain details for {domain_id} " + str(e))

    def update_domain(self, domain_id, config_body):
        """
        Function to update the domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :param config_body: JSON payload for the domain creation
        :type config_body: JSON dict
        :return: response dict
        """
        if domain_id is None or config_body is None:
            raise Exception("domain_id or config_body cannot be None")
        try:
            url_to_update_domain = url_builder.create_domain_url(self.client_config) + f"/{domain_id}"
            response = self.call_api("PATCH", url_to_update_domain,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error(f"{parsed_response}")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response, job_id=None)
        except Exception as e:
            self.logger.error(f"Failed to update the domain {domain_id}" + str(e))
            raise DomainError(f"Failed to update the domain {domain_id}" + str(e))

    def delete_domain(self, domain_id):
        """
        Function to delete the domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :return: response dict
        """
        if domain_id is None:
            raise Exception("domain_id cannot be None")
        url_to_delete_domain = url_builder.create_domain_url(self.client_config) + f"/{domain_id}"
        try:
            response = self.call_api("DELETE",
                                     url_to_delete_domain,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content,
            )
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error(f"Error in deleting the domain {domain_id}")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in deleting domain"),
                                                    response=parsed_response
                                                    )
        except Exception as e:
            self.logger.error(f"Error in deleting the domain {domain_id}")
            raise DomainError(f"Error in deleting the domain {domain_id} " + str(e))

    def get_sources_associated_with_domain(self, domain_id, params=None):
        """
        Function to get sources associated with the domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type params: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_sources_under_domain = url_builder.add_sources_to_domain_url(self.client_config,
                                                                                 domain_id) + IWUtils.get_query_params_string_from_dict(
            params=params)
        src_under_domain_list = []
        if domain_id is None:
            raise Exception("domain_id cannot be None")
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_sources_under_domain,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    src_under_domain_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
                response["result"] = src_under_domain_list
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    response="Failed to get sources associated with the domain"
                                                    )
        except Exception as e:
            self.logger.error("Error in listing sources under domain")
            raise DomainError("Error in listing sources under domain" + str(e))

    def add_source_to_domain(self, domain_id, config_body):
        """
        Function to add source to the domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :param config_body: JSON payload
        :type config_body: JSON dict
        config_body_example =
            {
             "entity_ids": ["33164bbaa811058db5c3941d"]
            }
        :return: response dict
        """
        if domain_id is None or config_body is None:
            raise Exception("domain_id or config_body cannot be None")
        try:
            url_to_add_source_domain = url_builder.add_sources_to_domain_url(self.client_config, domain_id)
            response = self.call_api("POST", url_to_add_source_domain,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error(f"Failed to add source to the domain {domain_id}")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get('details',
                                                                                   "Failed to add source to the domain"),
                                                    response=parsed_response)
        except Exception as e:
            raise DomainError(f"Failed to add source to the domain {domain_id}" + str(e))

    def remove_source_from_domain(self, domain_id, config_body):
        """
        Function to remove source from the domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :param config_body: JSON payload
        :type config_body: JSON dict

        config_body_example =
            {
             "entity_ids": ["33164bbaa811058db5c3941d"]
            }
        :return: response dict
        """
        if domain_id is None or config_body is None:
            raise Exception("domain_id or config_body cannot be None")
        try:
            url_to_delete_source_domain = url_builder.add_sources_to_domain_url(self.client_config, domain_id)
            response = self.call_api("DELETE", url_to_delete_source_domain,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error(f"Failed to remove source from the domain {domain_id}")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get('details',
                                                                                   "Failed to remove source from the domain"),
                                                    response=parsed_response)
        except Exception as e:
            raise DomainError(f"Failed to remove source from the domain {domain_id}" + str(e))

    def get_parent_entity_id(self, json_obj):
        """
        Function to get parent entity for any entity
        :param json_obj: Example: {"entity_id": workflow_id, "entity_type": "workflow"}
        :type json_obj: JSON dict
        :return: response dict
        """
        parent_entity_url = get_parent_entity_url(self.client_config)
        response = requests.request("GET", parent_entity_url, data=json_obj, verify=False)
        if response.status_code == 200 and len(response.json().get("result", [])) > 0:
            result = response.json().get("result", [])
            if len(result) > 0:
                domain_id = result[0]["entity_id"]
                self.logger.info("Domain ID is {} ".format(domain_id))
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response={"parent_entity_id": domain_id})
            else:
                self.logger.error("Failed to get parent entity id")
                self.logger.info("Domain ID is {} ".format(None))
                return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.GENERIC_ERROR,
                                                    response={"parent_entity_id": None})

    def get_domain_id(self, domain_name):
        """
        Function to get domain id using domain name
        :param domain_name: Name of the domain
        :type domain_name: String
        :return: Entity identifier of the domain
        """
        params = {"filter": {"name": domain_name}}
        url_to_list_domains = url_builder.create_domain_url(
            self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
        if domain_name is None:
            raise Exception("domain_name cannot be None")
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_domains,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if len(result) > 0:
                    domain_id = result[0]["id"]
                    return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                        response={"domain_id": domain_id})
                else:
                    self.logger.error(f"Failed to get domain id with name {domain_name}")
                    return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                        error_code=ErrorCode.GENERIC_ERROR,
                                                        response={"domain_id": None})
            else:
                self.logger.error("Failed to get domain id")
                return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.GENERIC_ERROR,
                                                    response={"domain_id": None})
        except Exception as e:
            self.logger.error("Error in finding domain id")
            raise DomainError("Error in finding domain id" + str(e))

    def get_pipeline_extensions_associated_with_domain(self, domain_id, params=None):
        """
        Function to get pipeline extensions associated with the domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type params: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_pl_extns_under_domain = url_builder.accessible_pipeline_extensions_url(self.client_config,
                                                                                           domain_id) + IWUtils.get_query_params_string_from_dict(
            params=params)
        pl_extn_under_domain_list = []
        if domain_id is None:
            raise Exception("domain_id cannot be None")
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_pl_extns_under_domain,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    pl_extn_under_domain_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
                response["result"] = pl_extn_under_domain_list
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    response="Failed to get pipeline extensions associated with the domain"
                                                    )
        except Exception as e:
            self.logger.error("Error in listing pipeline extensions under domain")
            raise DomainError("Error in listing pipeline extensions under domain" + str(e))

    def add_update_pipeline_extensions_to_domain(self, domain_id, config_body=None, action_type="create"):
        """
        Function to add pipeline extensions to the domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :param config_body: JSON payload
        :type config_body: JSON dict
        config_body_example =
            {
             "entity_ids": ["33164bbaa811058db5c3941d"]
            }
        :param action_type: Pass either create/update
        :return: response dict
        :return: response dict
        """
        if domain_id is None or config_body is None:
            raise Exception("domain_id or config_body cannot be None")
        try:
            if action_type.lower() == "create":
                request_type = "POST"
            else:
                request_type = "PUT"
            url_to_add_plextn_domain = url_builder.accessible_pipeline_extensions_url(self.client_config, domain_id)
            response = self.call_api(request_type, url_to_add_plextn_domain,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error(f"Failed to add/update pipeline extensions to the domain {domain_id}")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get('details',
                                                                                   "Failed to add/update pipeline extensions to the domain"),
                                                    response=parsed_response)
        except Exception as e:
            raise DomainError(f"Failed to add/update pipeline extensions to the domain {domain_id}" + str(e))

    def delete_pipeline_extensions_from_domain(self, domain_id, config_body=None):
        """
        Function to delete the pipeline extension from domain
        :param domain_id: Entity identifier for the domain
        :type domain_id: String
        :param config_body: JSON payload
        :type config_body: JSON dict
        config_body_example =
            {
             "entity_ids": ["33164bbaa811058db5c3941d"]
            }
        :return: response dict
        """
        if domain_id is None or config_body is None:
            raise Exception("domain_id or config_body cannot be None")
        url_to_delete_pl_extn = url_builder.accessible_pipeline_extensions_url(self.client_config, domain_id)
        try:
            response = self.call_api("DELETE",
                                     url_to_delete_pl_extn,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content,
            )
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error(f"Error in deleting the pipeline extension")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in deleting pipeline extension"),
                                                    response=parsed_response
                                                    )
        except Exception as e:
            self.logger.error(f"Error in deleting the pipeline extension")
            raise DomainError(f"Error in deleting the pipeline extension " + str(e))

    def modify_advanced_config_for_domain(self, domain_id, adv_config_body,
                                          action_type="update", key=None):
        """
        Function to add/update the adv config for the domain
        :param domain_id: Domain id to which the pipeline belongs to
        :type domain_id: String
        :param action_type: values can be either create/update. default update
        :type action_type: String
        :param adv_config_body: JSON dict
        adv_config_body_example = {
            "key" : "",
            "value": "",
            "description": "",
            "is_active": True
            }
        :param key: In case of update, name of the adv config to update
        :return: response dict
        """
        if None in {domain_id} or adv_config_body is None:
            raise Exception(f"domain_id and adv_config_body cannot be None")
        try:
            if action_type.lower() == "create":
                request_type = "POST"
                request_url = url_builder.advanced_config_domain_url(self.client_config, domain_id)
            else:
                request_type = "PUT"
                request_url = url_builder.advanced_config_domain_url(self.client_config, domain_id) + f"{key}"
            response = IWUtils.ejson_deserialize(
                self.call_api(request_type, request_url,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              adv_config_body).content)
            result = response.get('result', {})
            message = response.get('message', "")
            adv_config_id = result.get('id', None)
            if adv_config_id is not None:
                adv_config_id = str(adv_config_id)
                self.logger.info(
                    'Advanced Config has been created {id}.'.format(id=adv_config_id))
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=adv_config_id,
                                                    response=response)
            elif message == "Successfully updated Advance configuration":
                self.logger.info('Advanced Config has been updated')
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                self.logger.error(f'Failed to {action_type} advanced config for domain.')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc=f'Failed to {action_type} advanced config for domain',
                                                    response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to add adv config for domain.')
            raise DomainError('Error occurred while trying to add adv config for domain')

    def get_or_delete_advance_config_details_for_domain(self, domain_id, key,
                                                        action_type="get"):
        """
        Gets/Deletes advance configuration of domain
        :param domain_id: Entity identifier of the domain
        :type domain_id: String
        :param key: name of the advanced configurations
        :param action_type: values can be get/delete
        :return: response dict
        """
        try:
            if None in {domain_id, key}:
                raise Exception(f"domain_id, key cannot be None")
            request_type = "GET" if action_type.lower() == "get" else "DELETE"
            response = IWUtils.ejson_deserialize(
                self.call_api(request_type, url_builder.advanced_config_domain_url(
                    self.client_config, domain_id) + f"{key}", IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            message = response.get("message", "")
            if action_type.lower() == "delete" and message == "Successfully removed Advance configuration":
                self.logger.info(
                    'Successfully deleted the advanced config')
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response=response)

            if result.get('entity_id', None) is None:
                self.logger.error('Failed to find the advance config details')
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc='Failed to find the advance config details',
                                                    response=response)

            self.logger.info(
                'Successfully got the advanced config details.')
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, entity_id=result.get('entity_id', ''),
                                                response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(e))
            self.logger.exception('Error occurred while trying to get/delete adv config details.')
            raise DomainError('Error occurred while trying to get/delete adv config details.')



import copy
import json

import requests
import yaml
import configparser
from infoworks.sdk.url_builder import get_source_details_url
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.cicd.upload_configurations.update_configurations import InfoworksDynamicAccessNestedDict
from infoworks.sdk.cicd.upload_configurations.local_configurations import PRE_DEFINED_MAPPINGS

class CdataSource:
    def __init__(self):
        self.configuration_obj = None
        self.source_config_path = None
        self.environment_id = None
        self.storage_id = None
        self.secrets = None

    def set_variables(self, environment_id, storage_id, source_config_path, secrets=None, replace_words=""):
        self.storage_id = storage_id
        self.environment_id = environment_id
        self.source_config_path = source_config_path
        with open(self.source_config_path, 'r') as file:
            json_string = file.read()
            if replace_words != "":
                for key, value in [item.split("->") for item in replace_words.split(";")]:
                    json_string = json_string.replace(key, value)
        self.configuration_obj = IWUtils.ejson_deserialize(json_string)
        self.secrets = secrets

    def update_mappings_for_configurations(self, mappings):
        config = configparser.ConfigParser()
        config.read_dict(mappings)
        d = InfoworksDynamicAccessNestedDict(self.configuration_obj)
        for section in config.sections():
            if section in PRE_DEFINED_MAPPINGS:
                continue
            print("section:", section)
            try:
                final = d.setval(section.split("$"), dict(config.items(section)))
            except KeyError as e:
                pass
        self.configuration_obj = d.data
        if "configuration$source_configs$data_lake_schema" in config.sections():
            self.update_table_schema_and_database("target_schema",dict(config.items("configuration$source_configs$data_lake_schema")))
        if "configuration$source_configs$staging_schema_name" in config.sections():
            self.update_table_schema_and_database("stage_schema",dict(config.items("configuration$source_configs$staging_schema_name")))
        if "configuration$source_configs$target_database_name" in config.sections():
            self.update_table_schema_and_database("database",dict(config.items("configuration$source_configs$target_database_name")))

    def create_cdata_source(self, src_client_obj):
        data = self.configuration_obj["configuration"]["source_configs"]
        create_cdata_source_payload = data.copy()
        create_cdata_source_payload["data_lake_schema"]= data["data_lake_schema"] if "data_lake_schema" in data else ""
        create_cdata_source_payload["environment_id"] = self.environment_id
        create_cdata_source_payload["storage_id"] = self.storage_id
        create_cdata_source_payload["is_source_ingested"]= True
        src_create_response = src_client_obj.create_source(source_config=create_cdata_source_payload)
        if src_create_response["result"]["status"].upper() == "SUCCESS":
            source_id_created = src_create_response["result"]["source_id"]
            return source_id_created
        else:
            src_client_obj.logger.info('Cant create source {} '.format(data["name"]))
            src_client_obj.logger.info(f"Getting the existing SourceId with name {data['name']} if exists")
            filter_condition = IWUtils.ejson_serialize({"name": data['name']})
            source_detail_url = get_source_details_url(
                src_client_obj.client_config) + f"?filter={{filter_condition}}".format(
                filter_condition=filter_condition)
            response = requests.get(source_detail_url,
                                    headers={'Authorization': 'Bearer ' + src_client_obj.client_config['bearer_token'],
                                             'Content-Type': 'application/json'}, verify=False)
            if response.status_code == "406":
                headers = src_client_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + src_client_obj.client_config['bearer_token'],
                     'Content-Type': 'application/json'})
                response = requests.get(source_detail_url, headers=headers, verify=False)
            response = IWUtils.ejson_deserialize(response.content)
            if not response.get('result', None):
                src_client_obj.logger.error("Failed to make an api call to get source details")
                src_client_obj.logger.info(response)
            else:
                src_client_obj.logger.info(
                    f"Source Id with the same Source name {data['name']} : {response['result'][0]['id']}")
                return response['result'][0]['id']

    def configure_cdata_source_connection(self, src_client_obj, source_id, override_config_file=None,
                                          read_passwords_from_secrets=False, env_tag="", secret_type=""):
        source_configs = self.configuration_obj["configuration"]["source_configs"]
        src_name = str(source_configs["name"])
        connection_object = source_configs["connection"]
        connection_object["connection_mode"]="jdbc"
        if override_config_file is not None:
            with open(override_config_file) as file:
                information = yaml.load(file, Loader=yaml.FullLoader)
            if information["source_details"].get(src_name, None) is not None:
                override_keys = information["source_details"].get(src_name).keys()
                for key in override_keys:
                    connection_object[key] = information["source_details"][src_name][key]
        response = src_client_obj.configure_source_connection(source_id, connection_object=connection_object)
        if response["result"]["status"].upper() != "SUCCESS":
            src_client_obj.logger.info(f"Failed to configure the source {source_id} connection")
            src_client_obj.logger.info(response["result"])
            return "FAILED"
        else:
            src_client_obj.logger.info(response["result"])
            return "SUCCESS"

    def test_source_connection(self, src_client_obj, source_id):
        response = src_client_obj.source_test_connection_job_poll(source_id, poll_timeout=300,
                                                                  polling_frequency=15, retries=1)
        return response["result"]["status"].upper()

    def browse_source_tables(self, src_client_obj, source_id):
        filter_tables_properties = self.configuration_obj["filter_tables_properties"]
        response = src_client_obj.browse_source_tables(source_id, filter_tables_properties=filter_tables_properties,
                                                       poll_timeout=300, polling_frequency=15, retries=1)
        return response["result"]["status"].upper()

    def add_tables_to_source(self, src_client_obj, source_id):
        tables_already_added_in_source = src_client_obj.list_tables_in_source(source_id)["result"]["response"]
        tables_list = []
        tables = self.configuration_obj["configuration"]["table_configs"]
        if len(tables_already_added_in_source) > 0:
            for table in tables:
                if table["configuration"]["schema_name_at_source"] + "." + table["configuration"][
                    "name"] not in tables_already_added_in_source:
                    temp = {"table_name": table["configuration"]["name"],
                            "schema_name": table["configuration"]["schema_name_at_source"],
                            "table_type": table["entity_type"].upper(),
                            "target_table_name": table["configuration"]["configuration"]["target_table_name"],
                            "target_schema_name": table["configuration"]["configuration"]["target_schema_name"]}
                    if table["configuration"].get("catalog_name","")!="":
                        temp["catalog_name"]=table["configuration"]["catalog_name"]
                    tables_list.append(copy.deepcopy(temp))
                    src_client_obj.logger.info(
                        f"Adding table {temp['table_name']} to source {source_id} config payload")
        else:
            for table in tables:
                temp = {"table_name": table["configuration"]["name"],
                        "schema_name": table["configuration"]["schema_name_at_source"],
                        "table_type": table["entity_type"].upper(),
                        "target_table_name": table["configuration"]["configuration"]["target_table_name"],
                        "target_schema_name": table["configuration"]["configuration"]["target_schema_name"]}
                if table["configuration"].get("catalog_name", "") != "":
                    temp["catalog_name"] = table["configuration"]["catalog_name"]
                tables_list.append(copy.deepcopy(temp))
                src_client_obj.logger.info(f"Adding table {temp['table_name']} to source {source_id} config payload")
        response = src_client_obj.add_tables_to_source(source_id, tables_list)
        print(tables_list)
        return response["result"]["status"].upper()

    def configure_tables_and_tablegroups(self, src_client_obj, source_id, export_configuration_file=None,
                                         export_config_lookup=True, mappings=None, read_passwords_from_secrets=False,
                                         env_tag="", secret_type=""):
        if mappings is None:
            mappings = {}
        iw_mappings = self.configuration_obj["configuration"]["iw_mappings"]
        table_group_compute_mappings = mappings.get("table_group_compute_mappings", {})
        source_configs = self.configuration_obj["configuration"]["source_configs"]
        src_name = str(source_configs["name"])
        for item in iw_mappings:
            if item.get("entity_type", "") == "environment_compute_template":
                item["recommendation"]["compute_name"] = table_group_compute_mappings.get(
                    item["recommendation"]["compute_name"], item["recommendation"]["compute_name"])
        self.configuration_obj["configuration"]["iw_mappings"] = iw_mappings
        # Update the service account json file mappings if any
        if export_config_lookup and (export_configuration_file is not None or read_passwords_from_secrets):
            for table in self.configuration_obj["configuration"]["table_configs"]:
                # Check if there are any export configurations and passwords to replace
                if table.get("configuration", {}).get("export_configuration", None) is not None:
                    export_configs = table.get("configuration", {}).get("export_configuration")
                    target_type = export_configs.get("target_type", "").upper()
                    table_name = table["configuration"]["name"].upper()
                    override_keys = []
                    if export_configuration_file is not None:
                        with open(export_configuration_file) as file:
                            information = yaml.load(file, Loader=yaml.FullLoader)
                        if information["src_export_details"].get(src_name + "_" + table_name, None) is not None:
                            info_key = src_name + "_" + table_name
                            override_keys = information["src_export_details"].get(src_name + "_" + table_name,
                                                                                  {}).keys()
                        else:
                            info_key = src_name
                            override_keys = information["src_export_details"].get(src_name, {}).keys()
                        for key in override_keys:
                            table["configuration"]["export_configuration"]["connection"][key] = \
                                information["src_export_details"][info_key][key]
                    if target_type.upper() in ["SNOWFLAKE", "POSTGRES"]:
                        pass
                    elif target_type.upper() == "BIGQUERY":
                        if "server_path" not in override_keys:
                            server_path = table["configuration"]["export_configuration"].get("connection", {}).get(
                                "server_path", "")
                            if "gcp_details" in mappings:
                                server_path = mappings["gcp_details"].get("service_json_path")
                            if "service_json_mappings" in mappings:
                                server_path = mappings["service_json_mappings"].get(
                                    server_path.split("/")[-1],
                                    server_path)
                            table["configuration"]["export_configuration"]["connection"][
                                "server_path"] = server_path if server_path != "" else table["configuration"][
                                "export_configuration"].get("connection", {}).get(
                                "server_path", "")
                        table["configuration"]["export_configuration"]["connection"][
                            "upload_option"] = "serverLocation"

        response = src_client_obj.configure_tables_and_tablegroups(source_id, configuration_obj=self.configuration_obj[
            "configuration"])
        if response["result"]["status"].upper() != "SUCCESS":
            src_client_obj.logger.error("Failed to import the source {} (source config path : {})"
                                        .format(source_id, self.source_config_path))
            src_client_obj.logger.error(response.get("message", "") + "(source config path : {})"
                                        .format(self.source_config_path))
            return "FAILED"
        else:
            src_client_obj.logger.info(f"Successfully imported source configurations to {source_id}")
            return "SUCCESS"



import os

from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.cicd.download_configurations.utils import Utils


class DownloadPipeline(BaseClient):
    def __init__(self):
        super(DownloadPipeline, self).__init__()

    def cicd_get_pipelineconfig_dumps(self, pipeline_ids, config_file_dump_path, files_overwrite=True, serviceaccountemail="admin@infoworks.io", replace_words="",pipeline_grp_config=None):
        if not os.path.exists(os.path.join(config_file_dump_path, "modified_files")):
            os.makedirs(os.path.join(config_file_dump_path, "modified_files"))
        if not os.path.exists(os.path.join(config_file_dump_path, "pipeline")):
            os.makedirs(os.path.join(config_file_dump_path, "pipeline"))
        if not os.path.exists(os.path.join(config_file_dump_path, "pipeline_group")):
            os.makedirs(os.path.join(config_file_dump_path, "pipeline_group"))
        utils_obj = Utils(serviceaccountemail)
        target_file_path = os.path.join(os.path.join(config_file_dump_path, "modified_files"), "pipeline.csv")
        target_file_path_pipeline_groups = os.path.join(os.path.join(config_file_dump_path, "modified_files"), "pipeline_group.csv")
        if files_overwrite:
            open(target_file_path, 'w').close()
            open(target_file_path_pipeline_groups, 'w').close()
            mode = "a"
        else:
            mode = "a"
        f = open(target_file_path, mode)
        f_pipeline_group = open(target_file_path_pipeline_groups, mode)
        for pipeline_id in pipeline_ids:
            try:
                json_obj = {"entity_id": pipeline_id, "entity_type": "pipeline"}
                domain_id = utils_obj.get_domain_id(self, json_obj)
                if domain_id:
                    filename, configuration_obj = utils_obj.dump_to_file(self, "pipeline", domain_id, pipeline_id, replace_words, config_file_dump_path)
                    if filename is not None:
                        f.write(filename)
                        f.write("\n")
            except Exception as e:
                self.logger.error(f"Unable to export configurations for pipeline {pipeline_id} due to {str(e)}")
                print(f"Unable to export configurations for pipeline {pipeline_id} due to {str(e)}")
        if pipeline_grp_config is not None:
            pipeline_group_id = pipeline_grp_config["id"]
            json_obj = {"entity_id": pipeline_group_id, "entity_type": "pipeline_group"}
            domain_id = utils_obj.get_domain_id(self, json_obj)
            if domain_id:
                filename, configuration_obj = utils_obj.dump_to_file(self, "pipeline_group", domain_id, pipeline_group_id,
                                                                 replace_words, config_file_dump_path)
                if filename is not None:
                    f_pipeline_group.write(filename)
                    f_pipeline_group.write("\n")
        f.close()
        f_pipeline_group.close()


import os
import traceback

from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.cicd.download_configurations.utils import Utils


class DownloadSource(BaseClient):
    def __init__(self):
        super(DownloadSource, self).__init__()

    def cicd_get_sourceconfig_dumps(self, source_ids, config_file_dump_path, files_overwrite=True,
                                    serviceaccountemail="admin@infoworks.io",
                                    replace_words=""):
        # replace_words = "DEV->PROD;dev->prod"
        utils_obj = Utils(serviceaccountemail)
        if not os.path.exists(os.path.join(config_file_dump_path, "modified_files")):
            os.makedirs(os.path.join(config_file_dump_path, "modified_files"))
        if not os.path.exists(os.path.join(config_file_dump_path, "source")):
            os.makedirs(os.path.join(config_file_dump_path, "source"))
        target_file_path = os.path.join(os.path.join(config_file_dump_path, "modified_files"), "source.csv")
        if files_overwrite:
            open(target_file_path, 'w').close()
            mode = "a"
        else:
            mode = "a"
        f = open(target_file_path, mode)
        for source_id in source_ids:
            try:
                filename, configuration_obj = utils_obj.dump_to_file(self, "source", None,
                                                                     source_id, replace_words,
                                                                     config_file_dump_path)
                if filename is not None:
                    f.write(filename)
                    f.write("\n")
            except Exception as e:
                self.logger.error(f"Unable to export configurations for source {source_id} due to {str(e)}")
                print(f"Unable to export configurations for source {source_id} due to {str(e)}")
                print(traceback.format_exc())
        f.close()



import json
import re
from typing import List
import copy

def get_keys(d, curr_key=[]):
    if isinstance(d,str):
        return
    for k, v in d.items():
        if isinstance(v, dict):
            yield from get_keys(v, curr_key + [k])
        elif isinstance(v, list):
            for i in v:
                yield from get_keys(i, curr_key + [k])
        else:
            yield '.'.join(curr_key + [k])

class InfoworksDynamicAccessNestedDict:
    """Dynamically get/set nested dictionary keys of 'data' dict"""

    def __init__(self, data: dict):
        self.data = data
        self.flatten_json_list = [*get_keys(data)]

    def getval(self, keys: List):
        replacement_string = ".".join(keys)
        r = re.compile(replacement_string)
        matched_replacements = list(filter(r.match, self.flatten_json_list))
        data = self.data
        for item in matched_replacements:
            section_keys=item.split(".")
            for k in section_keys:
                try:
                    data = data[k]
                except KeyError as e:
                    print(f"Did not find the key {k} in artifact json")
                    return None
            return data

    def setval(self, keys: List,rename_dict) -> None:
        replacement_string = ".".join(keys)
        r=re.compile(replacement_string)
        matched_replacements = list(filter(r.match, self.flatten_json_list))
        for key in matched_replacements:
            section_keys = key.split(".")
            data = self.data
            lastkey = section_keys[-1]
            for k in section_keys[:-1]:
                data = data[k]
            try:
                if rename_dict.get(data[lastkey].lower(),{})!={} and data[lastkey].lower() in rename_dict.keys():
                    data[lastkey] = rename_dict.get(data[lastkey].lower())
            except KeyError as e:
                print(f"Could not find the key {lastkey} in artifact json.Skipping mappings updation...")




from infoworks.error import AdminError
from infoworks.sdk import url_builder
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.local_configurations import Response, ErrorCode
from infoworks.sdk.generic_response import GenericResponse
from infoworks.sdk.utils import IWUtils


class AdminClient(BaseClient):
    def __init__(self):
        super(AdminClient, self).__init__()
        self._datalineage = {"path": [], "dataflow_objects": [], "master_pipeline_ids": [],
                             "master_sourcetable_ids": []}

    def list_users(self, params=None):
        """
        Function to list the users
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_users = url_builder.list_users_url(self.client_config) \
                            + IWUtils.get_query_params_string_from_dict(params=params)

        users_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_users,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                while len(result) > 0:
                    users_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing users",
                                                            response=response
                                                            )

                response["result"] = users_list

            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing users")
            raise AdminError("Error in listing users" + str(e))

    def create_new_user(self, data=None, name=None, email=None, roles=None, password=None):
        """
        Function to create new user in Infoworks
        :param data: JSON Payload with user details
        :type data: JSON
        :param name: Name of the user
        :type name: String
        :param email: Email Id of the user
        :type email: String
        :param roles: Comma seperated list of roles
        :type roles: List
        :param password: Password of the user
        :type password: String

        example_data = {
            "profile": {
                "name": "iwx_sdk_user",
                "email": "iwx_sdk_user@infoworks.io",
                "needs_password_reset": False
            },
            "roles": ["modeller"],
            "password": ""
        }

        :return: response dict
        """
        if roles is None:
            roles = "modeller"
        if data is None:
            if None in {name, email, roles, password}:
                raise Exception(
                    f"User creation body (data) or all of the name, email, roles, password has to be passed")
            else:
                data = {
                    "profile": {
                        "name": name,
                        "email": email,
                        "needs_password_reset": False
                    },
                    "roles": roles.split(","),
                    "password": password
                }
        try:
            response = self.call_api("POST",
                                     url_builder.list_users_url(self.client_config),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=data)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in creating user ")
                                                    )
        except Exception as e:
            self.logger.error("Error in creating user " + str(e))
            raise AdminError("Error in creating user " + str(e))

    def update_the_user(self, user_id, data):
        """
        Function to update the user
        :param user_id: Entity identifier of the user
        :type user_id: String
        :param data: JSON Payload with user details
        :type data: JSON dict
        :return: response dict
        """
        try:
            response = self.call_api("PATCH",
                                     url_builder.list_users_url(self.client_config) + f"/{user_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=data)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in updating user ")
                                                    )
        except Exception as e:
            self.logger.error("Error in updating user " + str(e))
            raise AdminError("Error in updating user " + str(e))

    def delete_the_user(self, user_id):
        """
        Function to delete the user
        :param user_id: Entity identifier of the user
        :type user_id: String
        :return: response dict
        """
        try:
            response = self.call_api("DELETE",
                                     url_builder.list_users_url(self.client_config) + f"/{user_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                     )
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in deleting user ")
                                                    )
        except Exception as e:
            self.logger.error("Error in deleting user " + str(e))
            raise AdminError("Error in deleting user " + str(e))

    def get_user_details(self, user_id=None, params=None):
        """
        Function to get the user details
        :param user_id: Entity identifier of the user
        :type user_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type params: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        if user_id is not None:
            url_to_list_user_details = url_builder.list_users_url(self.client_config) + f"/{user_id}"
        else:
            url_to_list_user_details = url_builder.list_users_url(
                self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
        users_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_user_details,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", [])
                if user_id is not None:
                    users_list.extend([result])
                else:
                    while len(result) > 0:
                        users_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                response["result"] = users_list
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                self.logger.error("Failed to get user details")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    response="Failed to get user details"
                                                    )

        except Exception as e:
            self.logger.error("Error in listing user information")
            raise AdminError("Error in listing user information" + str(e))

    def add_domains_to_user(self, domain_id, user_id=None, user_email=None):
        """
        Function to make domains accessible to the user
        :param domain_id: Entity identifier for domain
        :type domain_id: String
        :param user_id: Entity identifier for user
        :type user_id: String
        :param user_email: email_id of the user
        :type user_email: String
        :return: response dict
        """
        status_flag = "failed"
        if user_id is None and user_email is not None:
            self.logger.info('Getting userID from given user email....')
            url_for_list_users = url_builder.list_users_url(self.client_config)
            filter_condition = IWUtils.ejson_serialize({"profile.email": user_email})
            get_userid_url = url_for_list_users + f"?filter={{filter_condition}}".format(
                filter_condition=filter_condition)
            if get_userid_url is not None:
                try:
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", get_userid_url,
                                      IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
                    self.logger.info(response)
                    if response is not None and "result" in response:
                        result = response.get("result", [])
                        if len(result) > 0:
                            user_id = result[0]["id"]
                except Exception as e:
                    self.logger.error('Couldnt get result for the user {}'.format(user_email))
                    self.logger.error(str(e))
                    raise AdminError('Couldnt get result for the user {}'.format(user_email))
                finally:
                    if user_id is None:
                        self.logger.error('Couldnt get result for the user {}'.format(user_email))
                        raise AdminError('Couldnt get result for the user {}'.format(user_email))
        response = ""
        if user_id is not None:
            url_for_adding_user_to_domain = url_builder.add_user_to_domain_url(self.client_config, user_id)
            self.logger.info('url for adding user to domain - ' + url_for_adding_user_to_domain)
            if url_for_adding_user_to_domain is not None:
                self.logger.info('Adding user {user} to domain {domain}'.format(user=user_id, domain=domain_id))
                try:
                    json_data = {"entity_ids": [str(domain_id)]}
                    response = IWUtils.ejson_deserialize(
                        self.call_api("POST",
                                      url_for_adding_user_to_domain,
                                      IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                      data=json_data).content)
                    if response.get('message', None) is not None:
                        self.logger.info(response.get('message'))
                        if response.get('message') == "Added Domain(s) to User":
                            status_flag = "success"
                    else:
                        self.logger.error('Error in adding user to domain - {}'.format(response))
                except Exception as e:
                    self.logger.error('Response from server while adding user to domain: {}'.format(user_id))
                    self.logger.error(str(e))
                    raise AdminError('Response from server while adding user to domain: {}'.format(user_id))
        if status_flag == "success":
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        else:
            return GenericResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                response=response)

    def alation_compatible_lineage_for_source(self, src_id, table_id):
        """
        Function to get the lineage of source
        :param src_id: Entity identifier for source
        :type src_id: String
        :param table_id: Entity identifier for table
        :type table_id: String
        :return: source lineage info
        """
        url_to_get_table_details = url_builder.get_table_configuration(self.client_config, src_id, table_id)
        response = self.call_api("GET", url_to_get_table_details,
                                 IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                 )
        parsed_response = IWUtils.ejson_deserialize(
            response.content)
        if response.status_code == 200 and "result" in parsed_response:
            result = parsed_response.get("result", {})
            target_schema_name = result["configuration"]["target_schema_name"]
            target_table_name = result["configuration"]["target_table_name"]
            orig_db_name = result.get("schema_name_at_source", "csv_source_db")
            orig_table_name = result.get("original_table_name", "")

            path_to_add = [[{"otype": "table",
                             "key": f"<DS_ID>.{orig_db_name}.{orig_table_name}"}],
                           [{"otype": "table",
                             "key": f"<DS_ID>.{target_schema_name}.{target_table_name}"}]]
            self._datalineage["path"].append(path_to_add[::])

            for column in result["columns"]:
                orig_name = column["original_name"]
                col_name = column["name"]
                path_to_add = [[{"otype": "column",
                                 "key": f"<DS_ID>.{orig_db_name}.{orig_table_name}.{orig_name}"}],
                               [{"otype": "column",
                                 "key": f"<DS_ID>.{target_schema_name}.{target_table_name}.{col_name}"}]]
                self._datalineage["path"].append(path_to_add[::])
        output = {"dataflow_objects": self._datalineage["dataflow_objects"], "paths": self._datalineage["path"]}
        return output

    def alation_compatible_lineage_for_pipeline(self, domain_id, pipeline_id, pipeline_version_id,
                                                pl_description=""):
        """
        Function to get the lineage of pipeline
        :param domain_id: Entity identifier for domain
        :type domain_id: String
        :param pipeline_id: Entity identifier for pipeline
        :type pipeline_id: String
        :param pipeline_version_id: Entity identifier for pipeline version
        :type pipeline_version_id: String
        :param pl_description: pipeline description
        :type pl_description: String
        :return: response dict
        """
        url_to_get_pipelineversion_details = url_builder.list_pipeline_versions_url(self.client_config, domain_id,
                                                                                    pipeline_id) + f"/{pipeline_version_id}"
        response = self.call_api("GET", url_to_get_pipelineversion_details,
                                 IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                 )
        parsed_response = IWUtils.ejson_deserialize(
            response.content)
        if response.status_code == 200 and len(parsed_response.get("result", {})) > 0:
            pipeline_details = parsed_response.get("result", {})
            # hash_object = hashlib.md5(f"{pipeline_id}_{pipeline_version_id}".encode())
            # hashcode = hash_object.hexdigest()
            hashcode = f"{pipeline_id}_{pipeline_version_id}"
            for node in pipeline_details["model"].get("nodes", []):
                req_dict = pipeline_details["model"]["nodes"][node]
                if req_dict['type'].upper() in ["TARGET", "BIGQUERY_TARGET"]:
                    src_node_details = []
                    if req_dict['type'].upper() == "TARGET":
                        target_schema_name = req_dict['properties'].get("target_schema_name", "")
                        target_table_name = req_dict['properties'].get("target_table_name", "")
                    else:
                        target_schema_name = req_dict['properties'].get("dataset_name", "")
                        target_table_name = req_dict['properties'].get("table_name", "")

                    output_columns = [i['name'] for i in req_dict["output_entities"]]
                    for column_name in output_columns:
                        # print(column_name)
                        if not column_name.upper().startswith("ZIW"):
                            url_to_get_pipeline_lineage = url_builder.get_pipeline_lineage_url(self.client_config,
                                                                                               domain_id,
                                                                                               pipeline_id,
                                                                                               pipeline_version_id,
                                                                                               column_name,
                                                                                               node)
                            response = self.call_api("GET", url_to_get_pipeline_lineage,
                                                     IWUtils.get_default_header_for_v3(
                                                         self.client_config['bearer_token'])
                                                     )
                            parsed_response = IWUtils.ejson_deserialize(
                                response.content)
                            if response.status_code == 200 and len(parsed_response.get("result", [])) > 0:
                                graph = parsed_response.get("result", [])
                                temp_output = []
                                source_node = graph[-1]
                                if source_node.get("node_type", "").upper() == "SOURCE_TABLE":
                                    pl_id = source_node.get("pipeline_id", None)
                                    if pl_id is not None and pl_id not in self._datalineage["master_pipeline_ids"]:
                                        self._datalineage["master_pipeline_ids"].append(pl_id)
                                    src_id = source_node.get("source_id", None)
                                    table_id = source_node.get("table_id", "")
                                    if src_id is not None and f"{src_id}:{table_id}" not in self._datalineage[
                                        "master_sourcetable_ids"]:
                                        self._datalineage["master_sourcetable_ids"].append(f"{src_id}:{table_id}")
                                    targetdl_table_name = source_node["target_table_name"]
                                    targetdl_schema_name = source_node["target_schema_name"]
                                    if f"<DS_ID>.{targetdl_schema_name}.{targetdl_table_name}" not in src_node_details:
                                        src_node_details.append(f"<DS_ID>.{targetdl_schema_name}.{targetdl_table_name}")
                                        path_to_add = [[{"otype": "table",
                                                         "key": f"<DS_ID>.{targetdl_schema_name}.{targetdl_table_name}"}],
                                                       [{"otype": "dataflow", "key": f"iwx_pipeline/{hashcode}"}],
                                                       [{"otype": "table",
                                                         "key": f"<DS_ID>.{target_schema_name}.{target_table_name}"}]]
                                        self._datalineage["path"].append(path_to_add[::])
                                    input_port_column = ""
                                    if source_node.get("input_port_column", {}) is not None:
                                        input_port_column = source_node.get("input_port_column", {}).get("name", "")
                                    key = f"<DS_ID>.{targetdl_schema_name}.{targetdl_table_name}.{input_port_column}"
                                    temp = [{"otype": "column", "key": key}]
                                    temp_output.append(temp[::])
                                temp_output.append([{"otype": "dataflow", "key": f"iwx_pipeline/{hashcode}"}])
                                key = f"<DS_ID>.{target_schema_name}.{target_table_name}.{column_name}"
                                temp = [{"otype": "column", "key": key}]
                                temp_output.append(temp[::])
                                self._datalineage["path"].append(temp_output[::])
                        else:
                            # ziw column direct connect to pipeline
                            path_to_add = [[{"otype": "dataflow", "key": f"iwx_pipeline/{hashcode}"}],
                                           [{"otype": "column",
                                             "key": f"<DS_ID>.{target_schema_name}.{target_table_name}.{column_name}"}]]
                            self._datalineage["path"].append(path_to_add[::])
            self._datalineage["dataflow_objects"].append(
                {"external_id": f"iwx_pipeline/{hashcode}", "content": pl_description})
            output = {"dataflow_objects": self._datalineage["dataflow_objects"], "paths": self._datalineage["path"]}
            return output

    def alation_compatible_lineage_for_iwx_metadata(self, domain_id, pipeline_id, sourcetable_ids=None):
        """
        Function to ger complete lineage info
        :param sourcetable_ids: Comma seperated src_id:table_id combination
        :type sourcetable_ids: Comma seperated src_id:table_id combination
        :param domain_id: Entity identifier for domain
        :type domain_id: String
        :param pipeline_id: Entity identifier for pipeline
        :type pipeline_id: String
        :return: lineage info
        """
        self._datalineage["master_pipeline_ids"].append(pipeline_id)
        # Check if there are any more master pipeline ids
        while len(self._datalineage["master_pipeline_ids"]) > 0:
            parent_pipeline_id = self._datalineage["master_pipeline_ids"].pop()
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_builder.get_pipeline_url(self.client_config, domain_id, parent_pipeline_id),
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                pipeline_version_id = response.get("result").get("active_version_id")
                pipeline_description = response.get("result").get("description", "")
                self.alation_compatible_lineage_for_pipeline(domain_id, parent_pipeline_id, pipeline_version_id,
                                                             pipeline_description)
        if sourcetable_ids is not None:
            self._datalineage["master_sourcetable_ids"].extend(sourcetable_ids.split(","))
        while len(self._datalineage["master_sourcetable_ids"]) > 0:
            src_table_id = self._datalineage["master_sourcetable_ids"].pop()
            src_id, table_id = src_table_id.split(":")
            self.alation_compatible_lineage_for_source(src_id, table_id)

    def get_environment_details(self, environment_id=None, params=None):
        """
        Function to get environment details
        :param environment_id: Entity identifier of the environment
        :type environment_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type params: Dict
        :return: Response Dict
        """
        if params is None and environment_id is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_environments = url_builder.get_environment_details(
            self.client_config, environment_id) + IWUtils.get_query_params_string_from_dict(params=params)
        env_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_environments,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", None)
                if result is None:
                    self.logger.error("Failed to get environment details for given id")
                    return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                        error_code=ErrorCode.USER_ERROR,
                                                        response=response,
                                                        error_desc="Failed to get environment details"
                                                        )
                initial_msg = response.get("message", "")
                if environment_id is not None:
                    env_details.extend([result])
                else:
                    while len(result) > 0:
                        env_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                response["result"] = env_details
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                self.logger.error("Failed to get environment details")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc="Failed to get environment details",
                                                    response=response
                                                    )
        except Exception as e:
            self.logger.error("Error in getting environment details " + str(e))
            raise AdminError("Error in getting environment details" + str(e))

    def get_storage_details(self, environment_id, storage_id=None, params=None):
        """
        Function to get storage details
        :param environment_id: Entity identifier of the environment
        :type environment_id: String
        :param storage_id: Entity identifier of the storage
        :type storage_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type params: Dict
        :return: Response Dict
        """
        if params is None and storage_id is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_storages = url_builder.get_environment_storage_details(
            self.client_config, environment_id, storage_id) + IWUtils.get_query_params_string_from_dict(params=params)
        storage_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_storages,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", None)
                if result is None:
                    self.logger.error("Failed to get environment storage details for given id")
                    return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                        error_code=ErrorCode.USER_ERROR,
                                                        response=response,
                                                        error_desc="Failed to get environment storage details"
                                                        )
                initial_msg = response.get("message", "")
                if storage_id is not None:
                    storage_details.extend([result])
                else:
                    while len(result) > 0:
                        storage_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                response["result"] = storage_details
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                self.logger.error("Failed to get storage details")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    response="Failed to get storage details"
                                                    )
        except Exception as e:
            self.logger.error("Error in getting storage details " + str(e))
            raise AdminError("Error in getting storage details" + str(e))

    def get_compute_template_details(self, environment_id, compute_id=None, is_interactive=False, params=None):
        """
         Function to get compute template details
         :param environment_id: Entity identifier of the environment
         :type environment_id: String
         :param compute_id: Entity identifier of compute cluster
         :type compute_id: String
         :param is_interactive: True/False. If True only the details of interactive clusters is fetched
         :type is_interactive: Boolean
         :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
         :type params: Dict
         :return: Response Dict
         """
        if params is None and compute_id is None:
            params = {"limit": 20, "offset": 0}
        if is_interactive:
            url_to_list_computes = url_builder.get_environment_interactive_compute_details(
                self.client_config, environment_id, compute_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
        else:
            url_to_list_computes = url_builder.get_environment_compute_details(
                self.client_config, environment_id, compute_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
        compute_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_computes,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None and "result" in response:
                result = response.get("result", None)
                if result is None:
                    self.logger.error("Failed to get environment compute details for given id")
                    return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                        error_code=ErrorCode.USER_ERROR,
                                                        response=response,
                                                        error_desc="Failed to get environment compute details"
                                                        )
                initial_msg = response.get("message", "")
                if compute_id is not None:
                    compute_details.extend([result])
                else:
                    while len(result) > 0:
                        compute_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                response["result"] = compute_details
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                self.logger.error("Failed to get compute template details")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    error_desc="Failed to get compute template details",
                                                    response=response
                                                    )
        except Exception as e:
            self.logger.error("Error in getting compute template details " + str(e))
            raise AdminError("Error in getting compute template details" + str(e))

    def get_environment_id_from_name(self, environment_name):
        """
        Function to return environment Id from name
        :param environment_name: Infoworks Environment Name
        :type environment_name: String
        :return: Response Dict
        """
        response = self.get_environment_details(environment_id=None, params={"filter": {"name": environment_name}})
        if response.get("result", {}).get("status", "") == Response.Status.SUCCESS:
            result = response.get("result", {}).get("response", {}).get("result", {})
            if len(result) > 0:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response={"environment_id": result[0]["id"]})
            else:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response={"environment_id": None}
                                                    )
        else:
            return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                error_code=ErrorCode.USER_ERROR,
                                                response=response.get("result", {}).get("response", {})
                                                )

    def get_environment_default_warehouse(self, environment_id):
        """
        Function to return default warehouse name
        :param environment_id: Infoworks Environment Id
        :type environment_id: String
        :return: Response Dict
        """
        response = self.get_environment_details(environment_id=environment_id, params=None)
        if response.get("result", {}).get("status", "") == Response.Status.SUCCESS:
            result = response.get("result", {}).get("response", {}).get("result", [])
            if len(result) > 0:
                for item in result[0]["data_warehouse_configuration"]["warehouse"]:
                    if item["is_default_warehouse"]:
                        return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                            response={"default_warehouse": item["name"]})
                else:
                    return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                        response={"default_warehouse":
                                                                      result[0]["data_warehouse_configuration"][
                                                                          "warehouse"][0]["name"]})
            else:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response=result
                                                    )
        else:
            return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                error_code=ErrorCode.USER_ERROR,
                                                response=response.get("result", {}).get("response", {})
                                                )

    def get_compute_id_from_name(self, environment_id, compute_name):
        """
        Function to return Environment Compute Id from name
        :param environment_id: Environment identifier
        :type environment_id: String
        :param compute_name: Environment Compute Name
        :type compute_name: String
        :return: Response Dict
        """
        response = self.get_compute_template_details(environment_id, compute_id=None, is_interactive=True,
                                                     params={"filter": {"name": compute_name}})
        if response.get("result", {}).get("status", "") == Response.Status.SUCCESS:
            result = response.get("result", {}).get("response", {}).get("result", [])
            if len(result) > 0:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response={"compute_id": result[0]["id"]})
            else:
                response_non_interactive = self.get_compute_template_details(environment_id, compute_id=None,
                                                                             is_interactive=False,
                                                                             params={"filter": {"name": compute_name}})
                result_non_interactive = response_non_interactive.get("result", {}).get("response", {}).get("result",
                                                                                                            [])
                if len(result_non_interactive) > 0:
                    return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                        response={"compute_id": result_non_interactive[0]["id"]})
                else:
                    return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                        response={"compute_id": None}
                                                        )
        else:
            return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                error_code=ErrorCode.USER_ERROR,
                                                response=response.get("result", {}).get("response", {})
                                                )

    def get_storage_id_from_name(self, environment_id, storage_name):
        """
        Function to return Environment Storage Id from name
        :param environment_id: Environment identifier
        :type environment_id: String
        :param storage_name: Environment Storage Name
        :type storage_name: String
        :return: Response Dict
        """
        response = self.get_storage_details(environment_id=environment_id,
                                            params={"filter": {"name": storage_name}})
        if response.get("result", {}).get("status", "") == Response.Status.SUCCESS:
            result = response.get("result", {}).get("response", {}).get("result", [])
            if len(result) > 0:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response={"storage_id": result[0]["id"]})
            else:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS,
                                                    response={"storage_id": None}
                                                    )
        else:
            return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                error_code=ErrorCode.USER_ERROR,
                                                response=response.get("result", {}).get("response", {})
                                                )

    def get_source_extension(self, extension_id):
        """
        Function to get a source extension details
        :param extension_id: ID of the source extension
        :type extension_id: String
        :return: Response Dict
        """
        try:
            response = self.call_api("GET",
                                     url_builder.create_source_extension_url(self.client_config) + f"/{extension_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                     )
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   f"Error getting source extension {extension_id}."),
                                                    response=parsed_response
                                                    )
        except Exception as e:
            self.logger.error(f"Error getting source extension {extension_id}.")
            raise AdminError(
                f"Error getting source extension {extension_id}." + str(e))

    def create_source_extension(self, data):
        """
        Function to create a source extension
        :param data: Source extension details body
        :type data: Dict
        :return: Response Dict
        """
        try:
            response = self.call_api("POST",
                                     url_builder.create_source_extension_url(self.client_config),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=data)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   f"Error creating source extension {data['name']}")
                                                    )
        except Exception as e:
            self.logger.error(f"Error creating source extension {data['name']}.Please do it by yourself from UI.")
            raise AdminError(
                f"Error creating source extension  {data['name']}.Please do it by yourself from UI." + str(e))

    def update_source_extension(self, extension_id, data=None):
        """
        Function to update a source extension
        :param extension_id: Id of the source extension to update
        :type extension_id: String
        :param data: JSON dict of the body to update
        :type data: JSON dict
        :return: Response Dict
        """
        if None in {extension_id} and data is None:
            self.logger.error("Extension id or data cannot be None")
            raise Exception("Extension id or data cannot be None")
        try:
            response = self.call_api("PUT",
                                     url_builder.create_source_extension_url(self.client_config) + f"/{extension_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                     , data=data)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   f"Error updating source extension {extension_id}."),
                                                    response=parsed_response
                                                    )
        except Exception as e:
            self.logger.error(f"Error updating source extension {extension_id}.")
            raise AdminError(
                f"Error updating source extension {extension_id}." + str(e))

    def delete_source_extension(self, extension_id):
        """
        Function to delete a source extension
        :param extension_id: ID of the source extension
        :type extension_id: String
        :return: Response Dict
        """
        try:
            response = self.call_api("DELETE",
                                     url_builder.create_source_extension_url(self.client_config) + f"/{extension_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                     )
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   f"Error deleting source extension {extension_id}."),
                                                    response=parsed_response
                                                    )
        except Exception as e:
            self.logger.error(f"Error deleting source extension {extension_id}.")
            raise AdminError(
                f"Error deleting source extension {extension_id}." + str(e))

    def create_data_connection(self, config_body):
        """
        Function to create data connection in the domain
        :param config_body: JSON payload
        :type config_body: JSON dict
        config_body_example = {
        "name": "Snowflake-connection",
        "type": "TARGET",
        "sub_type": "SNOWFLAKE",
        "properties": {
            "url": "https://account_name.snowflakecomputing.com",
            "account": "",
            "username": "ramesh",
            "password": {
                "password_type": "infoworks_managed"
                "secret_id": "6418304b7eeb1c40de2b6008" --> Needed if password_type is secret_store
            },
            "password": "", -->  Needed if the password_type is infoworks_managed
            "additional_params": "",
            "warehouse": "TEST_WH"
        }
        }
        :return: response dict
        """
        try:
            if config_body is None:
                raise Exception("config_body cannot be None/Null")
            create_data_connection_url = url_builder.create_data_connection(self.client_config)
            response = self.call_api("POST", create_data_connection_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error("Failed to create data connection")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Failed to create data connection"),
                                                    job_id=None, response=parsed_response)
        except Exception as e:
            raise AdminError(f"Failed to create data connection" + str(e))

    def update_data_connection(self, data_connection_id, config_body):
        """
        Function to create data connection in the domain
        :param data_connection_id: Entity identifier of the data connection id
        :param config_body: JSON payload
        :type config_body: JSON dict
        config_body_example = {
        "name": "Snowflake-connection",
        "type": "TARGET",
        "sub_type": "SNOWFLAKE",
        "properties": {
            "url": "https://account_name.snowflakecomputing.com",
            "account": "",
            "username": "ramesh",
            "password": {
                "password_type": "infoworks_managed",
                "secret_id": "6418304b7eeb1c40de2b6008" --> Needed if password_type is secret_store
            },
            "password": "", -->  Needed if the password_type is infoworks_managed
            "additional_params": "",
            "warehouse": "TEST_WH"
        }
        }
        :return: response dict
        """
        try:
            if config_body is None or data_connection_id is None:
                raise Exception("config_body and data_connection_id cannot be None/Null")
            create_data_connection_url = url_builder.create_data_connection(self.client_config)
            response = self.call_api("PUT", create_data_connection_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error("Failed to update the data connection")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Failed to update data connection"),
                                                    job_id=None, response=parsed_response)
        except Exception as e:
            raise AdminError(f"Failed to update the data connection" + str(e))

    def get_data_connection(self, data_connection_id=None, params=None):
        """
        Function to create data connection in the domain
        :param data_connection_id: id of dataconnection for which config has to be fetched
        :type data_connection_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}

        if data_connection_id is None:
            url_to_get_data_connection = url_builder.create_data_connection(
                self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
        else:
            url_to_get_data_connection = url_builder.create_data_connection(
                self.client_config) + f"/{data_connection_id}"
        dataconnection_list = []
        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_to_get_data_connection,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                if data_connection_id is not None:
                    dataconnection_list.extend(result)
                else:
                    while len(result) > 0:
                        dataconnection_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                response["result"] = dataconnection_list
                response["message"] = initial_msg
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                self.logger.error("Failed to get data connection details")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.USER_ERROR,
                                                    response="Failed to get data connection details"
                                                    )
        except Exception as e:
            raise AdminError(f"Failed to get data connection details" + str(e))

    def delete_data_connection(self, data_connection_id):
        """
        Function to delete data connection in the domain
        :param data_connection_id: Entity identifier for the data connection
        :type data_connection_id: String
        :return: response dict
        """
        try:
            if data_connection_id is None:
                raise Exception("data_connection_id cannot be None")
            create_data_connection_url = url_builder.create_data_connection(self.client_config).strip(
                "/") + f"/{data_connection_id}"
            response = self.call_api("DELETE", create_data_connection_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     )
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                self.logger.error("Failed to delete data connection")
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Failed to delete data connection"),
                                                    response=parsed_response)
        except Exception as e:
            raise AdminError(f"Failed in delete data connection" + str(e))

    def list_secret_stores(self, params=None):
        """
        Function to list the secret stores
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_secret_stores = url_builder.list_secret_stores_url(self.client_config) \
                                    + IWUtils.get_query_params_string_from_dict(params=params)

        secret_stores_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_secret_stores,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                while len(result) > 0:
                    secret_stores_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing secret store",
                                                            response=response
                                                            )

                response["result"] = secret_stores_list

            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing secret stores")
            raise AdminError("Error in listing secret store" + str(e))

    def create_secret_store(self, data=None):
        """
        Function to create new secret store in Infoworks
        :param data: JSON Payload with secret store details
        :type data: JSON
        example_data = {
            "name": "iwx-sp-secret-store",
            "service_type": "azure",
            "key_vault_uri": "https://iwazurecsdbkey.vault.azure.net/",
            "service_authentication": "64182fac35298f0d05a8b689",
        }
        :return: response dict
        """
        try:
            response = self.call_api("POST",
                                     url_builder.list_secret_stores_url(self.client_config),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=data)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in creating secret store ")
                                                    )
        except Exception as e:
            self.logger.error("Error in creating secret store " + str(e))
            raise AdminError("Error in creating secret store " + str(e))

    def get_secret_store_details(self, secret_store_id=None):
        """
        Function to get secret store details in Infoworks
        :param secret_store_id: secret store id in infoworks
        :type secret_store_id: String
        :return: response dict
        """
        try:
            response = self.call_api("GET",
                                     url_builder.list_secret_stores_url(self.client_config) + f"/{secret_store_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in getting secret store ")
                                                    )
        except Exception as e:
            self.logger.error("Error in getting secret store " + str(e))
            raise AdminError("Error in getting secret store " + str(e))

    def update_secret_store_details(self, secret_store_id=None, data=None):
        """
        Function to updates secret store details in Infoworks
        :param secret_store_id: secret store id in infoworks
        :type secret_store_id: String
        :param data: JSON Payload with secret store details
        :type data: JSON
        example_data = {
            "name": "iwx-sp-secret-store",
            "service_type": "azure",
            "key_vault_uri": "https://iwazurecsdbkey.vault.azure.net/",
            "service_authentication": "64182fac35298f0d05a8b689",
        }
        :return: response dict
        """
        try:
            response = self.call_api("PATCH",
                                     url_builder.list_secret_stores_url(self.client_config) + f"/{secret_store_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in updating secret store ")
                                                    )
        except Exception as e:
            self.logger.error("Error in updating secret store " + str(e))
            raise AdminError("Error in updating secret store " + str(e))

    def delete_secret_store(self, secret_store_id=None):
        """
        Function to delete secret store details in Infoworks
        :param secret_store_id: secret store id in infoworks
        :type secret_store_id: String
        :return: response dict
        """
        try:
            response = self.call_api("DELETE",
                                     url_builder.list_secret_stores_url(self.client_config) + f"/{secret_store_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in deleting secret store ")
                                                    )
        except Exception as e:
            self.logger.error("Error in deleting secret store " + str(e))
            raise AdminError("Error in deleting secret store " + str(e))

    def list_service_authentication(self, params=None):
        """
        Function to list the service authentication mechanisms in Infoworks
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_service_authentication = url_builder.list_service_authentication_url(self.client_config) \
                                             + IWUtils.get_query_params_string_from_dict(params=params)

        secret_stores_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_service_authentication,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                while len(result) > 0:
                    secret_stores_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing service authentication",
                                                            response=response
                                                            )

                response["result"] = secret_stores_list

            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing service authentication")
            raise AdminError("Error in listing service authentication" + str(e))

    def create_service_authentication(self, data=None):
        """
        Function to create new service authentication in Infoworks
        :param data: JSON Payload with service auth details
        :type data: JSON
        example_data = {
            "name": "iw-azure-cs-db-cluster-sp",
            "service_type": "azure",
            "authentication_type": "service_principal",
            "authentication_properties": {
                "subscription_id": "xxx",
                "client_id": "xxx",
                "object_id": "xxx",
                "tenant_id": "xxx"
            }
        }
        :return: response dict
        """
        try:
            response = self.call_api("POST",
                                     url_builder.list_service_authentication_url(self.client_config),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=data)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in creating service authentication mechanism ")
                                                    )
        except Exception as e:
            self.logger.error("Error in creating service authentication mechanism " + str(e))
            raise AdminError("Error in creating service authentication mechanism" + str(e))

    def get_service_authentication_details(self, service_auth_id=None):
        """
        Function to get service authentication in Infoworks
        :param service_auth_id: secret auth id in infoworks
        :type service_auth_id: String
        :return: response dict
        """
        try:
            response = self.call_api("GET",
                                     url_builder.list_service_authentication_url(
                                         self.client_config) + f"/{service_auth_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in getting service authentication ")
                                                    )
        except Exception as e:
            self.logger.error("Error in getting service authentication" + str(e))
            raise AdminError("Error in getting service authentication " + str(e))

    def update_service_authentication_details(self, service_auth_id=None, data=None):
        """
        Function to updates service auth details in Infoworks
        :param service_auth_id: service auth id in infoworks
        :type service_auth_id: String
        :param data: JSON Payload with service auth details
        :type data: JSON
        example_data = {
            "name": "iw-azure-cs-db-cluster-sp",
            "service_type": "azure",
            "authentication_type": "service_principal",
            "authentication_properties": {
                "subscription_id": "xxx",
                "client_id": "xxx",
                "object_id": "xxx",
                "tenant_id": "xxx"
            }
        }
        :return: response dict
        """
        try:
            response = self.call_api("PATCH",
                                     url_builder.list_service_authentication_url(
                                         self.client_config) + f"/{service_auth_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in updating service auth ")
                                                    )
        except Exception as e:
            self.logger.error("Error in updating service auth " + str(e))
            raise AdminError("Error in updating service auth " + str(e))

    def delete_service_authentication(self, service_auth_id=None):
        """
        Function to delete service authentication in Infoworks
        :param service_auth_id: secret auth id in infoworks
        :type service_auth_id: String
        :return: response dict
        """
        try:
            response = self.call_api("DELETE",
                                     url_builder.list_service_authentication_url(
                                         self.client_config) + f"/{service_auth_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in deleting service auth ")
                                                    )
        except Exception as e:
            self.logger.error("Error in deleting service auth " + str(e))
            raise AdminError("Error in deleting service auth " + str(e))

    def list_secrets(self, params=None):
        """
        Function to list the secrets in Infoworks
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_secrets = url_builder.list_secrets_url(self.client_config) \
                              + IWUtils.get_query_params_string_from_dict(params=params)

        secret_stores_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_secrets,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                while len(result) > 0:
                    secret_stores_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", None)
                    if result is None:
                        return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                            error_code=ErrorCode.GENERIC_ERROR,
                                                            error_desc="Error in listing secrets",
                                                            response=response
                                                            )

                response["result"] = secret_stores_list

            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error in listing secrets")
            raise AdminError("Error in listing secrets" + str(e))

    def create_secret(self, data=None):
        """
        Function to create new secret in Infoworks
        :param data: JSON Payload with secret details
        :type data: JSON
        example_data = {
            "name": "mongo-pg-pwd",
            "secret_store": "64182fd27eeb1c40de2b6007",
            "secret_name": "mongo-pg-pwd"
        }
        :return: response dict
        """
        try:
            response = self.call_api("POST",
                                     url_builder.list_secrets_url(self.client_config),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=data)
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in creating secret")
                                                    )
        except Exception as e:
            self.logger.error("Error in creating secret" + str(e))
            raise AdminError("Error in creating secret" + str(e))

    def get_secret_details(self, secret_id=None):
        """
        Function to get secret details in Infoworks
        :param secret_id: secret id in infoworks
        :type secret_id: String
        :return: response dict
        """
        try:
            response = self.call_api("GET",
                                     url_builder.list_secrets_url(self.client_config) + f"/{secret_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in getting secret details")
                                                    )
        except Exception as e:
            self.logger.error("Error in getting secret details" + str(e))
            raise AdminError("Error in getting secret details " + str(e))

    def update_secret_details(self, secret_id=None, data=None):
        """
        Function to updates secret details in Infoworks
        :param secret_id: secret id in infoworks
        :type secret_id: String
        :param data: JSON Payload with secret store details
        :type data: JSON
        example_data = {
            "name": "mongo-pg-pwd",
            "secret_store": "64182fd27eeb1c40de2b6007",
            "secret_name": "mongo-pg-pwd"
        }
        :return: response dict
        """
        try:
            response = self.call_api("PATCH",
                                     url_builder.list_secrets_url(self.client_config) + f"/{secret_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in updating secret ")
                                                    )
        except Exception as e:
            self.logger.error("Error in updating secret " + str(e))
            raise AdminError("Error in updating secret " + str(e))

    def delete_secret(self, secret_id=None):
        """
        Function to delete secret details in Infoworks
        :param secret_id: secret id in infoworks
        :type secret_id: String
        :return: response dict
        """
        try:
            response = self.call_api("DELETE",
                                     url_builder.list_secrets_url(self.client_config) + f"/{secret_id}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content)
            if response.status_code == 200:
                return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return GenericResponse.parse_result(status=Response.Status.FAILED,
                                                    error_code=ErrorCode.GENERIC_ERROR,
                                                    response=parsed_response,
                                                    error_desc=parsed_response.get("details",
                                                                                   "Error in deleting secret ")
                                                    )
        except Exception as e:
            self.logger.error("Error in deleting secret " + str(e))
            raise AdminError("Error in deleting secret " + str(e))